/**
  This file belong to the KMPlayer project, a movie player plugin for Konqueror
  Copyright (C) 2007  Koos Vriezen <koos.vriezen@gmail.com>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
**/

#include <config.h>

#include <math.h>
#include <cairo.h>
#include <cairo-xlib.h>
#include <pango/pangocairo.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
//#include <X11/extensions/Xdbe.h>

#include "kmplayertypes.h"
#include "kmplayer_smil.h"
#include "kmplayer.h"
#include "viewarea.h"
#include "kmplayer_rp.h"
#include "actor.h"

unsigned char * pixbufBits (GdkPixbuf *pixbuf, int & channels);

using namespace KMPlayer;

//-------------------------------------------------------------------------

namespace {
/**                                   a  b  0
 * Matrix for coordinate transforms   c  d  0
 *                                    tx ty 1     */
class KMPLAYER_NO_EXPORT Matrix {
    float a, b, c, d;
    Single tx, ty;
public:
    Matrix ();
    Matrix (const Matrix & matrix);
    Matrix (Single xoff, Single yoff, float xscale, float yscale);
    void getXY (Single & x, Single & y) const;
    void getXYWH (Single & x, Single & y, Single & w, Single & h) const;
    void invXYWH (Single & x, Single & y, Single & w, Single & h) const;
    void transform (const Matrix & matrix);
    void scale (float sx, float sy);
    void translate (Single x, Single y);
    // void rotate (float phi); // add this when needed
};

}

Matrix::Matrix () : a (1.0), b (0.0), c (0.0), d (1.0), tx (0), ty (0) {}

Matrix::Matrix (const Matrix & m)
 : a (m.a), b (m.b), c (m.c), d (m.d), tx (m.tx), ty (m.ty) {}

Matrix::Matrix (Single xoff, Single yoff, float xscale, float yscale)
 : a (xscale), b (0.0), c (0.0), d (yscale), tx (xoff), ty (yoff) {}

void Matrix::getXY (Single & x, Single & y) const {
    x = Single (x * a) + tx;
    y = Single (y * d) + ty;
}

void Matrix::getXYWH (Single & x, Single & y, Single & w, Single & h) const {
    getXY (x, y);
    w *= a;
    h *= d;
}

void Matrix::invXYWH (Single & x, Single & y, Single & w, Single & h) const {
    if (a > 0.00001 && d > 0.00001) {
        w /= a;
        h /= d;
        x = Single ((x - tx) / a);
        y = Single ((y - ty) / d);
    } else {
        warningLog () << "Not invering " << a << ", " << d << " scale" << endl;
    }
}

void Matrix::transform (const Matrix & matrix) {
    // TODO: rotate
    a *= matrix.a;
    d *= matrix.d;
    tx = Single (tx * matrix.a) + matrix.tx;
    ty = Single (ty * matrix.d) + matrix.ty;
}

void Matrix::scale (float sx, float sy) {
    a *= sx;
    d *= sy;
    tx *= sx;
    ty *= sy;
}

void Matrix::translate (Single x, Single y) {
    tx += x;
    ty += y;
}

//-----------------------------------------------------------------------------

void ImageData::copyImage (Surface *s, int w, int h, cairo_surface_t *similar) {
    cairo_surface_t *src_sf;

    unsigned char *bits = NULL;
    if (surface) {
        src_sf = surface;
    } else {
        const GPixbuf &gpb = (const GPixbuf) *image;
        int channels;
        unsigned char *bits = pixbufBits (gpb.pixbuf (), channels);
        has_alpha = 4 == channels;
        src_sf = cairo_image_surface_create_for_data (
                bits, has_alpha ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,
                width, height, width*4);
        if (flags & ImagePixmap && !(flags & ImageAnimated)) {
            surface = cairo_surface_create_similar (similar,
                    has_alpha ? CAIRO_CONTENT_COLOR_ALPHA : CAIRO_CONTENT_COLOR,
                    width, height);
            cairo_pattern_t *pat = cairo_pattern_create_for_surface (src_sf);
            cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
            cairo_t *cr = cairo_create (surface);
            cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
            cairo_set_source (cr, pat);
            cairo_paint (cr);
            cairo_destroy (cr);
            cairo_pattern_destroy (pat);
            cairo_surface_destroy (src_sf);
            g_free (bits);
            src_sf = surface;
            delete image;
            image = NULL;
        }
    }

    cairo_pattern_t *img_pat = cairo_pattern_create_for_surface (src_sf);
    cairo_pattern_set_extend (img_pat, CAIRO_EXTEND_NONE);
    if (w != width && h != height) {
        cairo_matrix_t mat;
        cairo_matrix_init_scale (&mat, 1.0 * width/w, 1.0 * height/h);
        cairo_pattern_set_matrix (img_pat, &mat);
    }
    if (!s->surface)
        s->surface = cairo_surface_create_similar (similar,
                has_alpha ?  CAIRO_CONTENT_COLOR_ALPHA : CAIRO_CONTENT_COLOR, w, h);
    cairo_t *cr = cairo_create (s->surface);
    cairo_set_source (cr, img_pat);
    if (!has_alpha)
        cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
    cairo_paint (cr);
    cairo_destroy (cr);

    cairo_pattern_destroy (img_pat);
    if (!surface) {
        cairo_surface_destroy (src_sf);
        g_free (bits);
    }
}

//-------------------------------------------------------------------------

namespace KMPlayer {

class KMPLAYER_NO_EXPORT ViewSurface : public Surface {
public:
    ViewSurface (ViewArea * widget);
    ViewSurface (ViewArea * widget, NodePtr owner, const SRect & rect);
    ~ViewSurface ();

    void clear () { m_first_child = 0L; }

    SurfacePtr createSurface (NodePtr owner, const SRect & rect);
    IRect toScreen (Single x, Single y, Single w, Single h);
    IRect clipToScreen (Single x, Single y, Single w, Single h);
    void resize (const SRect & rect);
    void repaint ();
    void repaint (const SRect &rect);

    NodePtrW current_video;
    ViewArea *view_widget;
};

} // namespace

KDE_NO_CDTOR_EXPORT ViewSurface::ViewSurface (ViewArea * widget)
  : Surface (NULL, SRect (0, 0, 342, 260)), //widget->width(),widget->height()))
    view_widget (widget)
{}

KDE_NO_CDTOR_EXPORT
ViewSurface::ViewSurface (ViewArea * widget, NodePtr owner, const SRect & rect)
  : Surface (owner, rect), view_widget (widget) {}

KDE_NO_CDTOR_EXPORT ViewSurface::~ViewSurface() {
    //debugLog () << "~ViewSurface" << endl;
}

KDE_NO_EXPORT
SurfacePtr ViewSurface::createSurface (NodePtr owner, const SRect & rect) {
    SurfacePtr surface = new ViewSurface (view_widget, owner, rect);
    appendChild (surface);
    return surface;
}

KDE_NO_EXPORT void ViewSurface::resize (const SRect &r) {
    bounds = r;
    /*if (rect == nrect)
        ;//return;
    SRect pr = rect.unite (nrect); // for repaint
    rect = nrect;*/
}

KDE_NO_EXPORT IRect ViewSurface::toScreen (Single x, Single y, Single w, Single h) {
    Matrix matrix (0, 0, xscale, yscale);
    matrix.translate (bounds.x (), bounds.y ());
    for (SurfacePtr s = parentNode(); s; s = s->parentNode()) {
        matrix.transform(Matrix (0, 0, s->xscale, s->yscale));
        matrix.translate (s->bounds.x (), s->bounds.y ());
    }
    matrix.getXYWH (x, y, w, h);
    return IRect (x, y, w, h);
}

KDE_NO_EXPORT IRect ViewSurface::clipToScreen (Single x, Single y, Single w, Single h) {
    Matrix m (0, 0, xscale, yscale);
    m.translate (bounds.x (), bounds.y ());
    m.getXYWH (x, y, w, h);
    SRect r = bounds.intersect (SRect (x, y, w, h));
    x= r.x (), y = r.y (), w = r.width (), h = r.height ();
    for (SurfacePtr s = parentNode (); s; s = s->parentNode ()) {
        m = Matrix (0, 0, s->xscale, s->yscale);
        m.translate (s->bounds.x (), s->bounds.y ());
        m.getXYWH (x, y, w, h);
        r = SRect (s->bounds).intersect (SRect (x, y, w, h));
        x= r.x (); y = r.y (); w = r.width (); h = r.height ();
    }
    return IRect (x, y, w, h);
}

KDE_NO_EXPORT void ViewSurface::repaint (const SRect &r) {
    markDirty ();
    view_widget->scheduleRepaint (clipToScreen (r.x (), r.y (), r.width (), r.height ()));
    //kdDebug() << "Surface::repaint x:" << (int)x << " y:" << (int)y << " w:" << (int)w << " h:" << (int)h << endl;
}

KDE_NO_EXPORT void ViewSurface::repaint () {
    markDirty ();
    view_widget->scheduleRepaint (clipToScreen (0, 0, bounds.width (), bounds.height ()));
}

//-------------------------------------------------------------------------

#define CAIRO_SET_SOURCE_RGB(cr,c)           \
    cairo_set_source_rgb ((cr),               \
            1.0 * (((c) >> 16) & 0xff) / 255, \
            1.0 * (((c) >> 8) & 0xff) / 255,  \
            1.0 * (((c)) & 0xff) / 255)

#define USE_IMG_SURFACE 0

class KMPLAYER_NO_EXPORT CairoPaintVisitor : public Visitor {
    IRect clip;
    ViewArea *view_widget;
    cairo_surface_t * cairo_surface;
    cairo_surface_t * cairo_surface_widget;
    Matrix matrix;
    // stack vars need for transitions
    SMIL::MediaType *cur_media;
    cairo_pattern_t * cur_pat;
    cairo_matrix_t cur_mat;
    float opacity;
    bool toplevel;

    void traverseRegion (Node *reg);
    void updateExternal (SMIL::MediaType *av, SurfacePtr s);
    void paint(SMIL::MediaType *, Surface *, int x, int y, const IRect &);
    void video (Mrl *mt, Surface *s);
public:
    cairo_t * cr;
    CairoPaintVisitor (ViewArea *v, cairo_surface_t * cs, Matrix m,
            const IRect & rect, bool toplevel=false);
    ~CairoPaintVisitor ();
    using Visitor::visit;
    void visit (Node * n);
    void visit (SMIL::Smil *);
    void visit (SMIL::Layout *);
    void visit (SMIL::RegionBase *);
    void visit (SMIL::Transition *);
    void visit (SMIL::ImageMediaType *);
    void visit (SMIL::TextMediaType *);
    void visit (SMIL::Brush *);
    void visit (SMIL::RefMediaType *);
    void visit (SMIL::AVMediaType *);
    void visit (RP::Imfl *);
    void visit (RP::Fill *);
    void visit (RP::Fadein *);
    void visit (RP::Fadeout *);
    void visit (RP::Crossfade *);
    void visit (RP::Wipe *);
    void visit (RP::ViewChange *);
};

KDE_NO_CDTOR_EXPORT
CairoPaintVisitor::CairoPaintVisitor (ViewArea *v, cairo_surface_t *cs,
        Matrix m, const IRect & rect, bool top)
 : view_widget (v), clip (rect),
#if USE_IMG_SURFACE
 cairo_surface_widget (cs), matrix (m), toplevel (top) {
    if (toplevel)
        cairo_surface = cairo_image_surface_create
            (CAIRO_FORMAT_RGB24, rect.w, rect.h);
    else
        cairo_surface = cs;
    cr = cairo_create (cairo_surface);
    if (toplevel) {
        cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
        cairo_set_tolerance (cr, 0.5 );
        cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
        cairo_rectangle (cr, 0, 0, rect.w, rect.h);
        cairo_fill (cr);
        cairo_translate (cr, -rect.x, -rect.y);
#else
# warning xrender
 cairo_surface (cs), matrix (m), toplevel (top) {
    cr = cairo_create (cs);
    if (toplevel) {
        cairo_rectangle (cr, rect.x, rect.y, rect.w, rect.h);
        cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
        cairo_set_tolerance (cr, 0.5 );
        cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
        //cairo_push_group (cr);
        cairo_set_source_rgb (cr, 32.0 / 255, 32.0 / 255, 48.0 / 255);
        cairo_rectangle (cr, rect.x, rect.y, rect.w, rect.h);
        cairo_fill (cr);
#endif
    } else {
        cairo_save (cr);
        cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
        cairo_rectangle (cr, rect.x, rect.y, rect.w, rect.h);
        cairo_fill (cr);
        cairo_restore (cr);
    }
}

KDE_NO_CDTOR_EXPORT CairoPaintVisitor::~CairoPaintVisitor () {
#if USE_IMG_SURFACE
    if (toplevel) {
        cairo_pattern_t *pat = cairo_pattern_create_for_surface (cairo_surface);
        cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
        cairo_matrix_t mat;
        cairo_matrix_init_translate (&mat, -clip.x, -clip.y);
        cairo_pattern_set_matrix (pat, &mat);
        cairo_t * wcr = cairo_create (cairo_surface_widget);
        cairo_set_source (wcr, pat);
        cairo_rectangle (wcr, clip.x, clip.y, clip.w, clip.h);
        cairo_fill (wcr);
        cairo_destroy (wcr);
        cairo_pattern_destroy (pat);
        cairo_surface_destroy (cairo_surface);
    }
#endif
    cairo_destroy (cr);
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (Node * n) {
    warningLog() << "Paint called on " << n->nodeName() << endl;
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::Smil *s) {
    if (s->active () && s->layout_node)
        s->layout_node->accept (this);
}

KDE_NO_EXPORT void CairoPaintVisitor::traverseRegion (Node *node) {
    // next visit receivers
    NodeRefListPtr nl = nodeMessageReceivers (node, MsgSurfaceAttach);
    if (nl) {
        for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ())
            if (c->data)
                c->data->accept (this);
    }
    // finally visit children, accounting for z-order FIXME optimize
    NodeRefList sorted;
    for (NodePtr n = node->firstChild (); n; n = n->nextSibling ()) {
        if (n->id != SMIL::id_node_region)
            continue;
        SMIL::Region * r = static_cast <SMIL::Region *> (n.ptr ());
        NodeRefItemPtr rn = sorted.first ();
        for (; rn; rn = rn->nextSibling ())
            if (r->z_order < convertNode <SMIL::Region> (rn->data)->z_order) {
                sorted.insertBefore (new NodeRefItem (n), rn);
                break;
            }
        if (!rn)
            sorted.append (new NodeRefItem (n));
    }
    for (NodeRefItemPtr r = sorted.first (); r; r = r->nextSibling ())
        r->data->accept (this);
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::Layout *reg) {
    //debugLog() << "Visit " << reg->nodeName() << endl;
    if (reg->root_layout) {
        Surface *s = (Surface *)reg->root_layout->message (MsgQueryRoleDisplay);
        if (s) {
            //cairo_save (cr);
            Matrix m = matrix;

            SRect rect = s->bounds;
            Single x, y, w = rect.width(), h = rect.height();
            matrix.getXYWH (x, y, w, h);

            IRect clip_save = clip;
            clip = clip.intersect (IRect (x, y, w, h));

            if (s->background_color & 0xff000000) {
                CAIRO_SET_SOURCE_RGB (cr, s->background_color);
                cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
                cairo_fill (cr);
            }

            matrix = Matrix (0, 0, s->xscale, s->yscale);
            matrix.transform (m);
            traverseRegion (reg->root_layout);
            traverseRegion (reg);
            //cairo_restore (cr);
            matrix = m;
            clip = clip_save;
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::RegionBase *reg) {
    Surface *s = (Surface *) reg->message (MsgQueryRoleDisplay);
    if (s) {
        SRect rect = s->bounds;

        Matrix m = matrix;
        Single x = rect.x(), y = rect.y(), w = rect.width(), h = rect.height();
        matrix.getXYWH (x, y, w, h);
        if (clip.intersect (IRect (x, y, w, h)).isEmpty ())
            return;
        matrix = Matrix (rect.x(), rect.y(), 1.0, 1.0);
        matrix.transform (m);
        IRect clip_save = clip;
        clip = clip.intersect (IRect (x, y, w, h));
        cairo_save (cr);
        ImageData *bg_img = reg->bg_image && !reg->bg_image->isEmpty()
            ? reg->bg_image->cached_img.ptr ()
            : NULL;
        if ((SMIL::RegionBase::ShowAlways == reg->show_background ||
                    reg->m_AttachedMediaTypes->first ()) &&
                (s->background_color & 0xff000000 || bg_img)) {
            cairo_save (cr);
            if (s->background_color & 0xff000000) {
                CAIRO_SET_SOURCE_RGB (cr, s->background_color);
                cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
                cairo_fill (cr);
            }
            if (bg_img) {
                Single x1, y1;
                Single w = bg_img->width;
                Single h = bg_img->height;
                matrix.getXYWH (x1, y1, w, h);
                if (!s->surface)
                    bg_img->copyImage (s, w, h, cairo_surface);
                if (bg_img->has_alpha)
                    cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
                cairo_pattern_t *pat = cairo_pattern_create_for_surface (s->surface);
                cairo_pattern_set_extend (pat, CAIRO_EXTEND_REPEAT);
                cairo_matrix_t mat;
                cairo_matrix_init_translate (&mat, (int) -x, (int) -y);
                cairo_pattern_set_matrix (pat, &mat);
                cairo_set_source (cr, pat);
                cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
                cairo_fill (cr);
                cairo_pattern_destroy (pat);
                if (bg_img->has_alpha)
                    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
            }
            cairo_restore (cr);
        }
        traverseRegion (reg);
        cairo_restore (cr);
        matrix = m;
        clip = clip_save;
    }
}

#define CAIRO_SET_PATTERN_COND(cr,pat,mat)                      \
    if (pat) {                                                  \
        cairo_pattern_set_extend (cur_pat, CAIRO_EXTEND_NONE);  \
        cairo_pattern_set_matrix (pat, &mat);                   \
        cairo_pattern_set_filter (pat, CAIRO_FILTER_FAST);      \
        cairo_set_source (cr, pat);                             \
    }

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::Transition *trans) {
    float perc = trans->start_progress + (trans->end_progress - trans->start_progress)*cur_media->trans_gain;
    if (cur_media->trans_out_active)
        perc = 1.0 - perc;
    if (SMIL::Transition::Fade == trans->type) {
        CAIRO_SET_PATTERN_COND(cr, cur_pat, cur_mat)
        cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
        opacity = perc;
    } else if (SMIL::Transition::BarWipe == trans->type) {
        IRect rect;
        if (SMIL::Transition::SubTopToBottom == trans->sub_type) {
            if (SMIL::Transition::dir_reverse == trans->direction) {
                int dy = (int) ((1.0 - perc) * clip.h);
                rect = IRect (clip.x, clip.y + dy, clip.w, clip.h - dy);
            } else {
                rect = IRect (clip.x, clip.y, clip.w, (int) (perc * clip.h));
            }
        } else {
            if (SMIL::Transition::dir_reverse == trans->direction) {
                int dx = (int) ((1.0 - perc) * clip.w);
                rect = IRect (clip.x + dx, clip.y, clip.w - dx, clip.h);
            } else {
                rect = IRect (clip.x, clip.y, (int) (perc * clip.w), clip.h);
            }
        }
        cairo_rectangle (cr, rect.x, rect.y, rect.w, rect.h);
        CAIRO_SET_PATTERN_COND(cr, cur_pat, cur_mat)
    } else if (SMIL::Transition::PushWipe == trans->type) {
        int dx = 0, dy = 0;
        if (SMIL::Transition::SubFromTop == trans->sub_type)
            dy = -(int) ((1.0 - perc) * clip.h);
        else if (SMIL::Transition::SubFromRight == trans->sub_type)
            dx = (int) ((1.0 - perc) * clip.w);
        else if (SMIL::Transition::SubFromBottom == trans->sub_type)
            dy = (int) ((1.0 - perc) * clip.h);
        else //if (SMIL::Transition::SubFromLeft == trans->sub_type)
            dx = -(int) ((1.0 - perc) * clip.w);
        cairo_matrix_translate (&cur_mat, -dx, -dy);
        IRect rect = clip.intersect (IRect (clip.x + dx, clip.y + dy, clip.w, clip.h));
        cairo_rectangle (cr, rect.x, rect.y, rect.w, rect.h);
        CAIRO_SET_PATTERN_COND(cr, cur_pat, cur_mat)
    } else if (SMIL::Transition::IrisWipe == trans->type) {
        CAIRO_SET_PATTERN_COND(cr, cur_pat, cur_mat)
        if (SMIL::Transition::SubDiamond == trans->sub_type) {
            cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
            cairo_clip (cr);
            int dx = (int) (perc * clip.w);
            int dy = (int) (perc * clip.h);
            int mx = clip.x + clip.w/2;
            int my = clip.y + clip.h/2;
            cairo_new_path (cr);
            cairo_move_to (cr, mx, my - dy);
            cairo_line_to (cr, mx + dx, my);
            cairo_line_to (cr, mx, my + dy);
            cairo_line_to (cr, mx - dx, my);
            cairo_close_path (cr);
        } else { // SubRectangle
            int dx = (int) (0.5 * (1 - perc) * clip.w);
            int dy = (int) (0.5 * (1 - perc) * clip.h);
            cairo_rectangle (cr, clip.x + dx, clip.y + dy,
                    clip.w - 2 * dx, clip.h -2 * dy);
        }
    } else if (SMIL::Transition::ClockWipe == trans->type) {
        cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
        cairo_clip (cr);
        int mx = clip.x + clip.w/2;
        int my = clip.y + clip.h/2;
        cairo_new_path (cr);
        cairo_move_to (cr, mx, my);
        float hw = 1.0 * clip.w/2;
        float hh = 1.0 * clip.h/2;
        float radius = sqrtf (hw * hw + hh * hh);
        float phi;
        switch (trans->sub_type) {
            case SMIL::Transition::SubClockwiseThree:
                phi = 0;
                break;
            case SMIL::Transition::SubClockwiseSix:
                phi = M_PI / 2;
                break;
            case SMIL::Transition::SubClockwiseNine:
                phi = M_PI;
                break;
            default: // Twelve
                phi = -M_PI / 2;
                break;
        }
        if (SMIL::Transition::dir_reverse == trans->direction)
            cairo_arc_negative (cr, mx, my, radius, phi, phi - 2 * M_PI * perc);
        else
            cairo_arc (cr, mx, my, radius, phi, phi + 2 * M_PI * perc);
        cairo_close_path (cr);
        CAIRO_SET_PATTERN_COND(cr, cur_pat, cur_mat)
    } else if (SMIL::Transition::BowTieWipe == trans->type) {
        cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
        cairo_clip (cr);
        int mx = clip.x + clip.w/2;
        int my = clip.y + clip.h/2;
        cairo_new_path (cr);
        cairo_move_to (cr, mx, my);
        float hw = 1.0 * clip.w/2;
        float hh = 1.0 * clip.h/2;
        float radius = sqrtf (hw * hw + hh * hh);
        float phi;
        switch (trans->sub_type) {
            case SMIL::Transition::SubHorizontal:
                phi = 0;
                break;
            default: // Vertical
                phi = -M_PI / 2;
                break;
        }
        float dphi = 0.5 * M_PI * perc;
        cairo_arc (cr, mx, my, radius, phi - dphi, phi + dphi);
        cairo_close_path (cr);
        cairo_new_sub_path (cr);
        cairo_move_to (cr, mx, my);
        if (SMIL::Transition::SubHorizontal == trans->sub_type)
            cairo_arc (cr, mx, my, radius, M_PI + phi - dphi, M_PI + phi +dphi);
        else
            cairo_arc (cr, mx, my, radius, -phi - dphi, -phi + dphi);
        cairo_close_path (cr);
        CAIRO_SET_PATTERN_COND(cr, cur_pat, cur_mat)
    } else if (SMIL::Transition::EllipseWipe == trans->type) {
        cairo_rectangle (cr, clip.x, clip.y, clip.w, clip.h);
        cairo_clip (cr);
        int mx = clip.x + clip.w/2;
        int my = clip.y + clip.h/2;
        float hw = (double) clip.w/2;
        float hh = (double) clip.h/2;
        float radius = sqrtf (hw * hw + hh * hh);
        cairo_save (cr);
        cairo_new_path (cr);
        cairo_translate (cr, (int) mx, (int) my);
        cairo_move_to (cr, - Single (radius), 0);
        if (SMIL::Transition::SubHorizontal == trans->sub_type)
            cairo_scale (cr, 1.0, 0.6);
        else if (SMIL::Transition::SubVertical == trans->sub_type)
            cairo_scale (cr, 0.6, 1.0);
        cairo_arc (cr, 0, 0, perc * radius, 0, 2 * M_PI);
        cairo_close_path (cr);
        cairo_restore (cr);
        CAIRO_SET_PATTERN_COND(cr, cur_pat, cur_mat)
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::video (Mrl *m, Surface *s) {
    if (m->actor &&
            ActorAgent::AudioVideo == m->actor->type ()) {
        AudioVideoActor *avm = static_cast<AudioVideoActor *> (m->actor);
        if (s && avm && strcmp (m->nodeName (), "audio")) {
            s->xscale = s->yscale = 1; // either scale width/heigt or use bounds
            view_widget->setAudioVideoGeometry (s->toScreen (
                        0, 0, s->bounds.width(), s->bounds.height ()), NULL);
        } else {
            view_widget->setAudioVideoGeometry (IRect (-60, -60, 50, 50), NULL);
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::RefMediaType *ref) {
    Surface *s = (Surface *) ref->message (MsgQueryRoleDisplay);
    if (s && ref->external_tree)
        updateExternal (ref, s);
    else
        video (ref, s);
}

KDE_NO_EXPORT void CairoPaintVisitor::paint (SMIL::MediaType *mt, Surface *s,
        int x, int y, const IRect &rect) {
    cairo_save (cr);
    opacity = 1.0;
    cairo_matrix_init_translate (&cur_mat, -x, -y);
    cur_pat = cairo_pattern_create_for_surface (s->surface);
    if (mt->active_trans) {
        IRect clip_save = clip;
        clip = rect;
        cur_media = mt;
        mt->active_trans->accept (this);
        clip = clip_save;
    } else {
        cairo_pattern_set_extend (cur_pat, CAIRO_EXTEND_NONE);
        cairo_pattern_set_matrix (cur_pat, &cur_mat);
        cairo_pattern_set_filter (cur_pat, CAIRO_FILTER_FAST);
        cairo_set_source (cr, cur_pat);
        cairo_rectangle (cr, rect.x, rect.y, rect.w, rect.h);
    }
    opacity *= mt->opacity / 100.0;
    if (opacity < 0.99) {
        cairo_operator_t op = cairo_get_operator (cr);
        cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
        cairo_clip (cr);
        cairo_paint_with_alpha (cr, opacity);
        cairo_set_operator (cr, op);
    } else {
        cairo_fill (cr);
    }
    cairo_pattern_destroy (cur_pat);
    cairo_restore (cr);
}

static Mrl *findActiveMrl (Node *n, bool *rp_or_smil) {
    Mrl *mrl = n->mrl ();
    if (mrl) {
        *rp_or_smil = (mrl->id >= SMIL::id_node_first &&
                mrl->id < SMIL::id_node_last) ||
            (mrl->id >= RP::id_node_first &&
             mrl->id < RP::id_node_last);
        if (*rp_or_smil ||
                (mrl->actor &&
                 ActorAgent::AudioVideo == mrl->actor->type ()))
            return mrl;
    }
    for (Node *c = n->firstChild ().ptr (); c; c = c->nextSibling ())
        if (c->active ()) {
            Mrl *m = findActiveMrl (c, rp_or_smil);
            if (m)
                return m;
        }
    return NULL;
}

KDE_NO_EXPORT
void CairoPaintVisitor::updateExternal (SMIL::MediaType *av, SurfacePtr s) {
    bool rp_or_smil = false;
    Mrl *ext_mrl = findActiveMrl (av->external_tree.ptr (), &rp_or_smil);
    if (!ext_mrl)
         return;
    if (!rp_or_smil) {
        video (ext_mrl, s.ptr ());
        return;
    }
    SRect rect = s->bounds;
    Single x = rect.x ();
    Single y = rect.y ();
    Single w = rect.width();
    Single h = rect.height();
    matrix.getXYWH (x, y, w, h);
    IRect clip_rect = clip.intersect (IRect (x, y, w, h));
    if (!clip_rect.isValid ())
        return;
    if (!s->surface || s->dirty) {
        Matrix m = matrix;
        m.translate (-x, -y);
        IRect r (clip_rect.x - (int) x - 1, clip_rect.y - (int) y - 1,
                clip_rect.w + 3, clip_rect.h + 3);
        if (!s->surface) {
            s->surface = cairo_surface_create_similar (cairo_surface,
                    CAIRO_CONTENT_COLOR_ALPHA, (int) w, (int) h);
            r = IRect (0, 0, w, h);
        }
        CairoPaintVisitor visitor (view_widget, s->surface, m, r);
        ext_mrl->accept (&visitor);
        s->dirty = false;
    }
    paint (av, s.ptr (), x, y, clip_rect);
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::AVMediaType *av) {
    Surface *s = (Surface *) av->message (MsgQueryRoleDisplay);
    if (s && av->external_tree)
        updateExternal (av, s);
    else
        video (av, s);
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::ImageMediaType * img) {
    //debugLog() << "Visit " << img->nodeName() << " " << img->src << endl;
    Surface *s = (Surface *) img->message (MsgQueryRoleDisplay);
    if (!s)
        return;
    if (img->external_tree) {
        updateExternal (img, s);
        return;
    }
    ImageActor *im = static_cast <ImageActor *> (img->actor);
    ImageData *id = im ? im->cached_img.ptr () : NULL;
    if (!id || im->isEmpty () || img->width <= 0 || img->height <= 0) {
        s->remove();
        return;
    }
    SRect rect = s->bounds;
    Single x = rect.x ();
    Single y = rect.y ();
    Single w = rect.width();
    Single h = rect.height();
    matrix.getXYWH (x, y, w, h);
    IRect clip_rect = clip.intersect (IRect (x, y, w, h));
    if (clip_rect.isEmpty ())
        return;
    if (!s->surface || s->dirty)
        id->copyImage (s, w, h, cairo_surface);
    if (id->has_alpha)
        cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
    paint (img, s, x, y, clip_rect);
    if (id->has_alpha)
        cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
    s->dirty = false;
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::TextMediaType * txt) {
    TextActor *tm = static_cast <TextActor *> (txt->actor);
    Surface *s = tm ? (Surface *) txt->message (MsgQueryRoleDisplay) : NULL;
    //debugLog() << "Visit " << txt->nodeName() << " " << td->text << endl;
    if (!s)
        return;
    if (!s->surface) {
        txt->width = txt->height = 0;
        s->bounds = txt->calculateBounds ();
    }
    SRect rect = s->bounds;
    Single x = rect.x (), y = rect.y(), w = rect.width(), h = rect.height();
    matrix.getXYWH (x, y, w, h);
    if (!s->surface) {
        /* QTextEdit * edit = new QTextEdit;
        edit->setReadOnly (true);
        edit->setHScrollBarMode (QScrollView::AlwaysOff);
        edit->setVScrollBarMode (QScrollView::AlwaysOff);
        edit->setFrameShape (QFrame::NoFrame);
        edit->setFrameShadow (QFrame::Plain);
        edit->setGeometry (0, 0, w, h);
        if (edit->length () == 0)
            edit->setText (text);
        if (w0 > 0)
            font.setPointSize (int (1.0 * w * font_size / w0));
        edit->setFont (font);
        QRect rect = p.clipRegion (QPainter::CoordPainter).boundingRect ();
        rect = rect.intersect (QRect (xoff, yoff, w, h));
        QPixmap pix = QPixmap::grabWidget (edit, rect.x () - (int) xoff,
                rect.y () - (int) yoff, rect.width (), rect.height ());*/

        cairo_surface_t *draw = cairo_surface_create_similar (cairo_surface,
                CAIRO_CONTENT_COLOR, (int) w, (int) h * 2); //FIXME
        cairo_t *cr_txt = cairo_create (draw);
        CAIRO_SET_SOURCE_RGB (cr_txt, txt->background_color);
        cairo_paint (cr_txt);

        CAIRO_SET_SOURCE_RGB (cr_txt, txt->font_color);
        PangoLayout *layout = pango_cairo_create_layout (cr_txt);
        pango_layout_set_width (layout, 1.0 * w * PANGO_SCALE);
        pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
        pango_layout_set_text (layout, (const char *)tm->text, -1);

        PangoFontDescription *desc = pango_font_description_new ();
        pango_font_description_set_family (desc, "Sans");
        pango_font_description_set_absolute_size (desc,
                PANGO_SCALE * (w * txt->font_size / rect.width ()));
        pango_layout_set_font_description (layout, desc);
        pango_font_description_free (desc);

        pango_cairo_show_layout (cr_txt, layout);
        int pxw, pxh;
        pango_layout_get_pixel_size (layout, &pxw, &pxh);
        g_object_unref (layout);
        cairo_destroy (cr_txt);

        cairo_pattern_t *pat = cairo_pattern_create_for_surface (draw);
        cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
        s->surface = cairo_surface_create_similar (cairo_surface,
                CAIRO_CONTENT_COLOR, pxw, pxh);
        cr_txt = cairo_create (s->surface);
        cairo_set_operator (cr_txt, CAIRO_OPERATOR_SOURCE);
        cairo_set_source (cr_txt, pat);
        cairo_paint (cr_txt);

        cairo_destroy (cr_txt);
        cairo_pattern_destroy (pat);
        cairo_surface_destroy (draw);

        // update bounds rect
        Single sx = x, sy = y, sw = pxw, sh = pxh;
        matrix.invXYWH (sx, sy, sw, sh);
        txt->width = sw;
        txt->height = sh;
        s->bounds = txt->calculateBounds ();

        // update coord. for painting below
        x = s->bounds.x ();
        y = s->bounds.y();
        w = s->bounds.width();
        h = s->bounds.height();
        matrix.getXYWH (x, y, w, h);
    }
    IRect clip_rect = clip.intersect (IRect (x, y, w, h));
    if (!clip_rect.isEmpty ())
        paint (txt, s, x, y, clip_rect);
    s->dirty = false;
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (SMIL::Brush * brush) {
    //debugLog() << "Visit " << brush->nodeName() << endl;
    Surface *s = (Surface *) brush->message (MsgQueryRoleDisplay);
    if (s) {
        cairo_save (cr);
        opacity = 1.0;
        SRect rect = s->bounds;
        Single x, y, w = rect.width(), h = rect.height();
        matrix.getXYWH (x, y, w, h);
        unsigned int color = Color (brush->param ("color")).rgb ();
        if (brush->active_trans) {
            cur_media = brush;
            cur_pat = NULL;
            brush->active_trans->accept (this);
        } else {
            cairo_rectangle (cr, (int) x, (int) y, (int) w, (int) h);
        }
        opacity *= brush->opacity / 100.0;
        if (opacity < 0.99) {
            cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
            cairo_set_source_rgba (cr,
                    1.0 * ((color >> 16) & 0xff) / 255,
                    1.0 * ((color >> 8) & 0xff) / 255,
                    1.0 * (color & 0xff) / 255,
                    opacity);
        } else {
            CAIRO_SET_SOURCE_RGB (cr, color);
        }
        cairo_fill (cr);
        if (opacity < 0.99)
            cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
        s->dirty = false;
        cairo_restore (cr);
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Imfl * imfl) {
    if (imfl->surface ()) {
        cairo_save (cr);
        Matrix m = matrix;
        Single x, y;
        Single w = imfl->rp_surface->bounds.width();
        Single h = imfl->rp_surface->bounds.height();
        matrix.getXYWH (x, y, w, h);
        cairo_rectangle (cr, x, y, w, h);
        //cairo_clip (cr);
        cairo_translate (cr, x, y);
        cairo_scale (cr, w/imfl->width, h/imfl->height);
        if (imfl->needs_scene_img)
            cairo_push_group (cr);
        for (NodePtr n = imfl->firstChild (); n; n = n->nextSibling ())
            if (n->state >= Node::state_began &&
                    n->state < Node::state_deactivated) {
                RP::TimingsBase * tb = convertNode<RP::TimingsBase>(n);
                switch (n->id) {
                    case RP::id_node_viewchange:
                        if (!(int)tb->srcw)
                            tb->srcw = imfl->width;
                        if (!(int)tb->srch)
                            tb->srch = imfl->height;
                        // fall through
                    case RP::id_node_crossfade:
                    case RP::id_node_fadein:
                    case RP::id_node_fadeout:
                    case RP::id_node_fill:
                    case RP::id_node_wipe:
                        if (!(int)tb->w)
                            tb->w = imfl->width;
                        if (!(int)tb->h)
                            tb->h = imfl->height;
                        n->accept (this);
                        break;
                }
            }
        if (imfl->needs_scene_img) {
            cairo_pattern_t * pat = cairo_pop_group (cr);
            cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
            cairo_set_source (cr, pat);
            cairo_paint (cr);
            cairo_pattern_destroy (pat);
        }
        cairo_restore (cr);
        matrix = m;
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Fill * fi) {
    //kdDebug() << "Visit " << fi->nodeName() << endl;
    CAIRO_SET_SOURCE_RGB (cr, fi->color);
    if ((int)fi->w && (int)fi->h) {
        cairo_rectangle (cr, fi->x, fi->y, fi->w, fi->h);
        cairo_fill (cr);
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Fadein * fi) {
    //debugLog() << "Visit " << fi->nodeName() << endl;
    if (fi->target && fi->target->id == RP::id_node_image) {
        RP::Image *img = convertNode <RP::Image> (fi->target);
        ImageActor *im = img ? static_cast <ImageActor *> (img->actor) : NULL;
        if (im && img->surface ()) {
            Single sx = fi->srcx, sy = fi->srcy, sw = fi->srcw, sh = fi->srch;
            if (!(int)sw)
                sw = img->width;
            if (!(int)sh)
                sh = img->height;
            if ((int)fi->w && (int)fi->h && (int)sw && (int)sh) {
                if (!img->img_surface->surface)
                    im->cached_img->copyImage (img->img_surface,
                            img->width, img->height, cairo_surface);
                cairo_matrix_t matrix;
                cairo_matrix_init_identity (&matrix);
                float scalex = 1.0 * sw / fi->w;
                float scaley = 1.0 * sh / fi->h;
                cairo_matrix_scale (&matrix, scalex, scaley);
                cairo_matrix_translate (&matrix,
                        1.0*sx/scalex - (double)fi->x,
                        1.0*sy/scaley - (double)fi->y);
                cairo_save (cr);
                cairo_rectangle (cr, fi->x, fi->y, fi->w, fi->h);
                cairo_pattern_t *pat = cairo_pattern_create_for_surface (img->img_surface->surface);
                cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
                cairo_pattern_set_matrix (pat, &matrix);
                cairo_set_source (cr, pat);
                cairo_clip (cr);
                cairo_paint_with_alpha (cr, 1.0 * fi->progress / 100);
                cairo_restore (cr);
                cairo_pattern_destroy (pat);
            }
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Fadeout * fo) {
    //kdDebug() << "Visit " << fo->nodeName() << endl;
    if (fo->progress > 0) {
        CAIRO_SET_SOURCE_RGB (cr, fo->to_color);
        if ((int)fo->w && (int)fo->h) {
            cairo_save (cr);
            cairo_rectangle (cr, fo->x, fo->y, fo->w, fo->h);
            cairo_clip (cr);
            cairo_paint_with_alpha (cr, 1.0 * fo->progress / 100);
            cairo_restore (cr);
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Crossfade * cf) {
    //debugLog() << "Visit " << cf->nodeName() << endl;
    if (cf->target && cf->target->id == RP::id_node_image) {
        RP::Image *img = convertNode <RP::Image> (cf->target);
        ImageActor *im = img ? static_cast <ImageActor *> (img->actor) : NULL;
        if (im && img->surface ()) {
            Single sx = cf->srcx, sy = cf->srcy, sw = cf->srcw, sh = cf->srch;
            if (!(int)sw)
                sw = img->width;
            if (!(int)sh)
                sh = img->height;
            if ((int)cf->w && (int)cf->h && (int)sw && (int)sh) {
                if (!img->img_surface->surface)
                    im->cached_img->copyImage (img->img_surface,
                            img->width, img->height, cairo_surface);
                cairo_save (cr);
                cairo_matrix_t matrix;
                cairo_matrix_init_identity (&matrix);
                float scalex = 1.0 * sw / cf->w;
                float scaley = 1.0 * sh / cf->h;
                cairo_matrix_scale (&matrix, scalex, scaley);
                cairo_matrix_translate (&matrix,
                        1.0*sx/scalex - (double)cf->x,
                        1.0*sy/scaley - (double)cf->y);
                cairo_rectangle (cr, cf->x, cf->y, cf->w, cf->h);
                cairo_pattern_t *pat = cairo_pattern_create_for_surface (img->img_surface->surface);
                cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
                cairo_pattern_set_matrix (pat, &matrix);
                cairo_set_source (cr, pat);
                cairo_clip (cr);
                cairo_paint_with_alpha (cr, 1.0 * cf->progress / 100);
                cairo_restore (cr);
                cairo_pattern_destroy (pat);
            }
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::Wipe * wipe) {
    //debugLog() << "Visit " << wipe->nodeName() << endl;
    if (wipe->target && wipe->target->id == RP::id_node_image) {
        RP::Image *img = convertNode <RP::Image> (wipe->target);
        ImageActor *im = img ? static_cast <ImageActor *> (img->actor) : NULL;
        if (im && img->surface ()) {
            Single x = wipe->x, y = wipe->y;
            Single tx = x, ty = y;
            Single w = wipe->w, h = wipe->h;
            Single sx = wipe->srcx, sy = wipe->srcy, sw = wipe->srcw, sh = wipe->srch;
            if (!(int)sw)
                sw = img->width;
            if (!(int)sh)
                sh = img->height;
            if (wipe->direction == RP::Wipe::dir_right) {
                Single dx = w * 1.0 * wipe->progress / 100;
                tx = x -w + dx;
                w = dx;
            } else if (wipe->direction == RP::Wipe::dir_left) {
                Single dx = w * 1.0 * wipe->progress / 100;
                tx = x + w - dx;
                x = tx;
                w = dx;
            } else if (wipe->direction == RP::Wipe::dir_down) {
                Single dy = h * 1.0 * wipe->progress / 100;
                ty = y - h + dy;
                h = dy;
            } else if (wipe->direction == RP::Wipe::dir_up) {
                Single dy = h * 1.0 * wipe->progress / 100;
                ty = y + h - dy;
                y = ty;
                h = dy;
            }

            if ((int)w && (int)h) {
                if (!img->img_surface->surface)
                    im->cached_img->copyImage (img->img_surface,
                            img->width, img->height, cairo_surface);
                cairo_matrix_t matrix;
                cairo_matrix_init_identity (&matrix);
                float scalex = 1.0 * sw / wipe->w;
                float scaley = 1.0 * sh / wipe->h;
                cairo_matrix_scale (&matrix, scalex, scaley);
                cairo_matrix_translate (&matrix,
                        1.0*sx/scalex - (double)tx,
                        1.0*sy/scaley - (double)ty);
                cairo_pattern_t *pat = cairo_pattern_create_for_surface (img->img_surface->surface);
                cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
                cairo_pattern_set_matrix (pat, &matrix);
                cairo_set_source (cr, pat);
                cairo_rectangle (cr, x, y, w, h);
                cairo_fill (cr);
                cairo_pattern_destroy (pat);
            }
        }
    }
}

KDE_NO_EXPORT void CairoPaintVisitor::visit (RP::ViewChange * vc) {
    //kdDebug() << "Visit " << vc->nodeName() << endl;
    if (vc->unfinished () || vc->progress < 100) {
        cairo_pattern_t * pat = cairo_pop_group (cr); // from imfl
        cairo_pattern_set_extend (pat, CAIRO_EXTEND_NONE);
        cairo_push_group (cr);
        cairo_save (cr);
        cairo_set_source (cr, pat);
        cairo_paint (cr);
        if ((int)vc->w && (int)vc->h && (int)vc->srcw && (int)vc->srch) {
            cairo_matrix_t matrix;
            cairo_matrix_init_identity (&matrix);
            float scalex = 1.0 * vc->srcw / vc->w;
            float scaley = 1.0 * vc->srch / vc->h;
            cairo_matrix_scale (&matrix, scalex, scaley);
            cairo_matrix_translate (&matrix,
                    1.0*vc->srcx/scalex - (double)vc->x,
                    1.0*vc->srcy/scaley - (double)vc->y);
            cairo_pattern_set_matrix (pat, &matrix);
            cairo_set_source (cr, pat);
            cairo_rectangle (cr, vc->x, vc->y, vc->w, vc->h);
            cairo_fill (cr);
        }
        cairo_pattern_destroy (pat);
        cairo_restore (cr);
    }
}

//-----------------------------------------------------------------------------

namespace KMPlayer {

class KMPLAYER_NO_EXPORT MouseVisitor : public Visitor {
    Matrix matrix;
    NodePtr node;
    MessageType event;
    int x, y;
    bool handled;
    bool bubble_up;
public:
    MouseVisitor (MessageType evt, int x, int y);
    KDE_NO_CDTOR_EXPORT ~MouseVisitor () {}
    using Visitor::visit;
    void visit (Node * n);
    void visit (Element *);
    void visit (SMIL::Smil *);
    void visit (SMIL::Layout *);
    void visit (SMIL::RegionBase *);
    void visit (SMIL::MediaType * n);
    void visit (SMIL::Anchor *);
    void visit (SMIL::Area *);

};

} // namespace

KDE_NO_CDTOR_EXPORT
MouseVisitor::MouseVisitor (MessageType evt, int a, int b)
    : event (evt), x (a), y (b), handled (false), bubble_up (false) {
}

KDE_NO_EXPORT void MouseVisitor::visit (Node * n) {
    debugLog () << "Mouse event ignored for " << n->nodeName () << endl;
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Smil *s) {
    if (s->active () && s->layout_node)
        s->layout_node->accept (this);
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Layout * layout) {
    Surface *s = (Surface *) layout->root_layout->message (MsgQueryRoleDisplay);
    if (s) {
        Matrix m = matrix;
        SRect rect = s->bounds;
        matrix = Matrix (rect.x(), rect.y(), s->xscale, s->yscale);
        matrix.transform (m);

        NodePtr node_save = node;
        node = layout;
        for (NodePtr r = layout->firstChild (); r; r = r->nextSibling ()) {
            if (r->id == SMIL::id_node_region)
                r->accept (this);
            if (!node->active ())
                break;
        }
        node = node_save;

        matrix = m;
    }
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::RegionBase *region) {
    if (region->message (MsgQueryRoleDisplay)) {
        SRect rect = region->region_surface->bounds;
        Single rx = rect.x(), ry = rect.y(), rw = rect.width(), rh = rect.height();
        matrix.getXYWH (rx, ry, rw, rh);
        handled = false;
        bool inside = x > rx && x < rx+rw && y > ry && y< ry+rh;
        if (!inside && (event == MsgEventClicked || !region->has_mouse))
            return;
        Matrix m = matrix;
        matrix = Matrix (rect.x(), rect.y(), 1.0, 1.0);
        matrix.transform (m);
        bubble_up = false;

        bool child_handled = false;
        if (inside)
            for (NodePtr r = region->firstChild (); r; r = r->nextSibling ()) {
                r->accept (this);
                child_handled |= handled;
                if (!node->active ())
                    break;
            }
        child_handled &= !bubble_up;
        bubble_up = false;

        MessageType saved_event = event;
        if (node->active ()) {
            bool notify_receivers = !child_handled;
            bool pass_event = !child_handled;
            if (event == MsgEventPointerMoved) {
                pass_event = true; // always pass move events
                if (region->has_mouse && (!inside || child_handled)) {
                    notify_receivers = true;
                    region->has_mouse = false;
                    event = MsgEventPointerOutBounds;
                } else if (inside && !child_handled && !region->has_mouse) {
                    notify_receivers = true;
                    region->has_mouse = true;
                    event = MsgEventPointerInBounds;
                }
            }// else // MsgEventClicked
            if (notify_receivers) {
                Posting mouse_event (region, event);
                region->deliver (event, &mouse_event);
            }
            if (pass_event) {
                NodeRefList *nl = nodeMessageReceivers (region, MsgSurfaceAttach);
                if (nl) {
                    for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ()) {
                        if (c->data)
                            c->data->accept (this);
                        if (!node->active ())
                            break;
                    }
                }
            }
        }
        event = saved_event;
        handled = inside;
        matrix = m;
    }
}

static void followLink (SMIL::LinkingBase * link) {
    debugLog () << "link to " << link->href << " clicked" << endl;
    if (link->href.startsWith ("#")) {
        SMIL::Smil * s = SMIL::Smil::findSmilNode (link);
        if (s)
            s->jump (link->href.mid (1));
        else
            errorLog() << "In document jumps smil not found" << endl;
    } else {
        PlayListNotify *notify = link->document ()->notify_listener;
        if (notify && !link->target.isEmpty ()) {
             notify->openUrl (link->href, link->target, String ());
        } else {
            NodePtr n = link;
            for (NodePtr p = link->parentNode (); p; p = p->parentNode ()) {
                if (n->mrl () && n->mrl ()->opener == p) {
                    p->setState (Node::state_deferred);
                    p->mrl ()->setParam (StringPool::attr_src, link->href, 0L);
                    p->activate ();
                    break;
                }
                n = p;
            }
        }
    }
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Anchor * anchor) {
    if (event == MsgEventPointerMoved)
        ;// set cursor to hand
    else if (event == MsgEventClicked)
        followLink (anchor);
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::Area * area) {
    NodePtr n = area->parentNode ();
    if (n->id >= SMIL::id_node_first_mediatype &&
            n->id < SMIL::id_node_last_mediatype) {
        SMIL::MediaType * mt = convertNode <SMIL::MediaType> (n);
        Surface *s = (Surface *) mt->message (MsgQueryRoleDisplay);
        if (s) {
            SRect rect = s->bounds;
            Single x1 = rect.x (), x2 = rect.y ();
            Single w = rect.width (), h = rect.height ();
            matrix.getXYWH (x1, x2, w, h);
            if (area->nr_coords > 1) {
                Single left = area->coords[0].size (rect.width ());
                Single top = area->coords[1].size (rect.height ());
                matrix.getXY (left, top);
                if (x < left || x > left + w || y < top || y > top + h)
                    return;
                if (area->nr_coords > 3) {
                    Single right = area->coords[2].size (rect.width ());
                    Single bottom = area->coords[3].size (rect.height ());
                    matrix.getXY (right, bottom);
                    if (x > right || y > bottom)
                        return;
                }
            }
            if (event == MsgEventPointerMoved)
                ;//cursor.setShape (Qt::PointingHandCursor);
            else {
                NodeRefList *nl = nodeMessageReceivers (area, event);
                if (nl)
                    for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ()) {
                        if (c->data)
                            c->data->accept (this);
                        if (!node->active ())
                            return;
                    }
                if (event == MsgEventClicked && !area->href.isEmpty ())
                    followLink (area);
            }
        }
    }
}

KDE_NO_EXPORT void MouseVisitor::visit (Element *elm) {
    Runtime *rt = (Runtime *) elm->message (MsgQueryRoleTiming);
    if (rt)
        rt->processEvent (event);
}

KDE_NO_EXPORT void MouseVisitor::visit (SMIL::MediaType *mt) {
    if (mt->sensitivity == SMIL::MediaType::sens_transparent) {
        bubble_up = true;
        return;
    }
    Surface *s = (Surface *) mt->message (MsgQueryRoleDisplay);
    if (!s)
        return;
    if (s->node && s->node.ptr () != mt) {
        s->node->accept (this);
        return;
    }
    SRect rect = s->bounds;
    Single rx = rect.x(), ry = rect.y(), rw = rect.width(), rh = rect.height();
    matrix.getXYWH (rx, ry, rw, rh);
    const bool inside = x > rx && x < rx+rw && y > ry && y< ry+rh;
    if (!inside && (event == MsgEventClicked || inside == mt->has_mouse))
        return;
    mt->has_mouse = inside;

    NodeRefList *nl = nodeMessageReceivers (mt,
            event == MsgEventPointerMoved ? MsgSurfaceAttach : event);
    if (nl)
        for (NodeRefItemPtr c = nl->first(); c; c = c->nextSibling ()) {
            if (c->data && c->data.ptr () != mt)
                c->data->accept (this);
            if (!node->active ())
                return;
        }
    if (event != MsgEventPointerMoved)
        visit (static_cast <Element *> (mt));
    if (event != MsgEventPointerInBounds && event != MsgEventPointerOutBounds) {
      SMIL::RegionBase *r=convertNode<SMIL::RegionBase>(mt->region_node);
      if (r && r->message (MsgQueryRoleDisplay) &&
              r->id != SMIL::id_node_smil &&
              r->region_surface->node && r != r->region_surface->node.ptr ())
          return r->region_surface->node->accept (this);
    }
}

//-----------------------------------------------------------------------------

namespace KMPlayer {
class KMPLAYER_NO_EXPORT ViewAreaPrivate {
public:
    ViewAreaPrivate (ViewArea *v)
        : m_view_area (v), dpy (NULL), backing_store (0) {}

    ~ViewAreaPrivate() {
        destroyBackingStore ();
    }

    Display *display () {
        if (!dpy)
            dpy = gdk_x11_display_get_xdisplay (gtk_widget_get_display (
                        m_view_area->viewWidget ()));
        return dpy;
    }
    void resizeSurface (Surface *s) {
        int w = m_view_area->width ();
        int h = m_view_area->height ();
        if (s->surface && (w != width || h != height)) {
                //cairo_xlib_surface_set_size (s->surface, w, h);
            Window wid = gdk_x11_drawable_get_xid (
                    GDK_DRAWABLE (m_view_area->viewWidget ()->window));
            XFreePixmap (display (), backing_store);
            backing_store = XCreatePixmap (dpy, wid, w, h,
                    gdk_drawable_get_depth (
                        GDK_DRAWABLE (m_view_area->viewWidget ()->window)));
            cairo_xlib_surface_set_drawable(s->surface, backing_store, w,h);
            width = w;
            height = h;
        }
    }
    cairo_surface_t *createSurface (int w, int h) {
        Window wid = gdk_x11_drawable_get_xid (
                GDK_DRAWABLE (m_view_area->viewWidget ()->window));
        Display *d = display ();
        destroyBackingStore ();
        width = w;
        height = h;
        backing_store = XCreatePixmap (d, wid, w, h, gdk_drawable_get_depth (
                    GDK_DRAWABLE (m_view_area->viewWidget ()->window)));
        return cairo_xlib_surface_create (d, backing_store,
                DefaultVisual (d, DefaultScreen (d)), w, h);
        /*return cairo_xlib_surface_create_with_xrender_format (
            QX11Info::display (),
            id,
            DefaultScreenOfDisplay (QX11Info::display ()),
            XRenderFindVisualFormat (QX11Info::display (),
                DefaultVisual (QX11Info::display (),
                    DefaultScreen (QX11Info::display ()))),
            w, h);*/
    }
    void swapBuffer (int sx, int sy, int sw, int sh, int dx, int dy) {
        Window wid = gdk_x11_drawable_get_xid (
                GDK_DRAWABLE (m_view_area->viewWidget ()->window));
        GC gc = XCreateGC (display (), backing_store, 0, NULL);
        XCopyArea (dpy, backing_store, wid, gc, sx, sy, sw, sh, dx, dy);
        XFreeGC (dpy, gc);
        XFlush (dpy);
    }
    void destroyBackingStore () {
        if (backing_store)
            XFreePixmap (display (), backing_store);
        backing_store = 0;
    }
    ViewArea *m_view_area;
    Display *dpy;
    Drawable backing_store;
    int width;
    int height;
};

class KMPLAYER_NO_EXPORT RepaintUpdater {
public:
    RepaintUpdater (Node *n, RepaintUpdater *nx) : node (n), next (nx) {}

    NodePtrW node;
    RepaintUpdater *next;
};

}

//-----------------------------------------------------------------------------

KDE_NO_CDTOR_EXPORT ViewArea::ViewArea ()
 : d (new ViewAreaPrivate (this)),
   m_updaters (NULL),
   aspect (0.0),
   widget (0L),
   video (0L),
   surface (new ViewSurface (this)),
   m_repaint_timer (0),
   m_fullscreen (false),
   m_event_filter_on (false),
   m_updaters_enabled (true) {
}

KDE_NO_CDTOR_EXPORT ViewArea::~ViewArea () {
    reset ();
    delete d;
}

KDE_NO_EXPORT void ViewArea::reset () {
    if (m_repaint_timer)
        g_source_remove (m_repaint_timer);
    m_repaint_timer = 0;
    while (m_updaters) {
        RepaintUpdater *tmp = m_updaters;
        m_updaters = m_updaters->next;
        delete tmp;
    }
    m_updaters = NULL;
    aspect = 0,0;
}

KDE_NO_EXPORT void ViewArea::mousePressEvent (int x, int y) {
    if (surface->node) {
        MouseVisitor visitor (MsgEventClicked, x, y);
        surface->node->accept (&visitor);
    }
}

KDE_NO_EXPORT void ViewArea::mouseDoubleClickEvent () {
}

KDE_NO_EXPORT void ViewArea::syncVisual (const IRect & rect) {
    int ex = rect.x;
    if (ex > 0)
        ex--;
    int ey = rect.y;
    if (ey > 0)
        ey--;
    int ew = rect.w + 2;
    int eh = rect.h + 2;
    if (!surface->surface)
        surface->surface = d->createSurface (width (), height ());
    {
        CairoPaintVisitor visitor (this, surface->surface,
                Matrix (surface->bounds.x(), surface->bounds.y(), 1.0, 1.0),
                IRect (ex, ey, ew, eh), true);
        if (surface->node)
            surface->node->accept (&visitor);
    }
    cairo_surface_flush (surface->surface);
    d->swapBuffer (ex, ey, ew, eh, ex, ey);

    m_repaint_rect = IRect ();
}

KDE_NO_EXPORT void ViewArea::paintEvent (const IRect & rect) {
    scheduleRepaint (rect);
}

KDE_NO_EXPORT void ViewArea::updateSurfaceBounds () {
    Single x, y, w = width (), h = height ();
    surface->resize (SRect (x, y, w, h));
    Mrl *mrl = surface->node ? surface->node->mrl () : NULL;
    if (w > 0 && h > 0 &&
            mrl && mrl->width > 0 && mrl->height > 0) {
        double wasp = (double) w / h;
        double masp = (double) mrl->width / mrl->height;
        if (wasp > masp) {
            Single tmp = w;
            w = masp * h;
            x += (tmp - w) / 2;
        } else {
            Single tmp = h;
            h = Single (w / masp);
            y += (tmp - h) / 2;
        }
        surface->xscale = 1.0 * w / mrl->width;
        surface->yscale = 1.0 * h / mrl->height;
    } else {
        surface->xscale = 1.0;
        surface->yscale = 1.0;
    }
    surface->bounds = SRect (x, y, w, h);
    scheduleRepaint (IRect (0, 0, width (), height ()));
}

KDE_NO_EXPORT void ViewArea::resizeEvent () {
    if (!video)
        return;
    Single x, y, w = width (), h = height ();
    if (w <= 0 || h <= 0)
        return;
    if (surface->node) {
        NodePtr n = surface->node;
        surface = new ViewSurface (this);
        surface->node = n;
    }
    updateSurfaceBounds ();
    d->resizeSurface (surface.ptr ());
    if (video_node && Mrl::SingleMode == video_node->mrl ()->view_mode)
        setAudioVideoGeometry (IRect (x, y, w, h), 0L);
}

KDE_NO_EXPORT
void ViewArea::setAudioVideoGeometry (const IRect &rect, unsigned int * bg_color) {
    int x = rect.x, y = rect.y, w = rect.w, h = rect.h;
    if (!surface->node && aspect > 0.01) {
        int w1 = w;
        int h1 = h;
        w = int (h * aspect);
        if (w > w1) {
            h = int (w1 / aspect);
            w = w1;
        }
        x += (w1 - w) / 2;
        y += (h1 - h) / 2;
    }
    Color color (bg_color ? (int)*bg_color : 0);
    Rect wrect = Rect (x, y, w, h);
    if (m_av_geometry != wrect) {
        if (m_av_geometry.width () > 0 && m_av_geometry.height () > 0)
            scheduleRepaint (IRect (m_av_geometry.x (), m_av_geometry.y (),
                    m_av_geometry.width (), m_av_geometry.height ()));
        m_av_geometry = Rect (x, y, w, h);
        debugLog() << "ViewArea::setAudioVideoGeometry " << x << "," << y << " "
            << w << "x" << h << endl;
        gdk_window_move_resize (video, x, y, w, h);
        gdk_window_set_background (video, (GdkColor *)(GColor)color);
        gdk_window_clear (video);
    }
}

KDE_NO_EXPORT void ViewArea::setAudioVideoNode (NodePtr n) {
    video_node = n;
    m_av_geometry = Rect ();
    if (!n || Mrl::WindowMode == n->mrl ()->view_mode)
        setAudioVideoGeometry (IRect (-100, -100, 50, 50), 0L);
}

KDE_NO_EXPORT
void ViewArea::setAspect (float a) {
    aspect = a;
    /*if (aspect > 0.01) {
        int w, h, x, y;
        gdk_drawable_get_size (video, &w, &h);
        gdk_window_get_position (video, &x, &y);
        scheduleRepaint (x, y, w, h);
        int w1 = int (h * aspect);
        int h1 = h;
        if (w1 > w) {
            h1 = int (w / aspect);
            w1 = w;
        }
        int xoff = (w - w1) / 2;
        int yoff = (h - h1) / 2;
        if (xoff > 0 && yoff > 0)
            gdk_window_move_resize (video, x + xoff, y + yoff, w1, h1);
    }*/
}

KDE_NO_EXPORT Surface *ViewArea::getSurface (Mrl *mrl) {
    static_cast <ViewSurface *> (surface.ptr ())->clear ();
    surface->node = mrl;
    //m_view->viewer()->resetBackgroundColor ();
    if (mrl) {
        updateSurfaceBounds ();
        return surface.ptr ();
    } else if (surface->surface) {
        cairo_surface_destroy (surface->surface);
        surface->surface = 0L;
        d->destroyBackingStore ();
    }
    scheduleRepaint (IRect (0, 0, width (), height ()));
    return 0L;
}

KDE_NO_EXPORT void ViewArea::scheduleRepaint (const IRect &rect) {
    if (m_repaint_timer)
        m_repaint_rect = m_repaint_rect.unite (rect);
    else {
        m_repaint_rect = rect;
        m_repaint_timer = g_timeout_add (50, cb_scheduledPaint, this);
    }
}

KDE_NO_EXPORT void ViewArea::addUpdater (Node *node) {
    m_updaters = new RepaintUpdater (node, m_updaters);
    if (m_updaters_enabled && !m_repaint_timer)
        m_repaint_timer = g_timeout_add (50, cb_scheduledPaint, this);
}

KDE_NO_EXPORT void ViewArea::removeUpdater (Node *node) {
    RepaintUpdater *prev = NULL;
    for (RepaintUpdater *r = m_updaters; r; r = r->next) {
        if (r->node.ptr () == node) {
            if (prev)
                prev->next = r->next;
            else
                m_updaters = r->next;
            delete r;
            break;
        }
        prev = r;
    }
    if (m_repaint_timer &&
            (!m_updaters_enabled || !m_updaters) &&
            m_repaint_rect.isEmpty ()) {
        g_source_remove (m_repaint_timer);
        m_repaint_timer = 0;
    }
}

static RepaintUpdater *getFirstUpdater (RepaintUpdater *updaters) {
    for (RepaintUpdater *r = updaters; r; r = updaters) {
        if (updaters->node)
            return updaters;
        updaters = r->next;
        delete r;
    }
    return NULL;
}

static void propagateUpdatersEvent (RepaintUpdater *updaters, void *event) {
    for (RepaintUpdater *r = updaters; r; ) {
        RepaintUpdater *next = r->next;
        if (r->node)
            r->node->message (MsgSurfaceUpdate, event); // may call removeUpdater()
        r = next;
    }
}

KDE_NO_EXPORT
void ViewArea::enableUpdaters (bool enable, unsigned int skip) {
    m_updaters_enabled = enable;
    m_updaters = getFirstUpdater (m_updaters);
    if (enable && m_updaters) {
        UpdateEvent event (m_updaters->node->document (), skip);
        propagateUpdatersEvent (m_updaters, &event);
        if (!m_repaint_timer)
            m_repaint_timer = g_timeout_add (50, cb_scheduledPaint, this);;
    } else if (!enable && m_repaint_timer && m_repaint_rect.isEmpty ()) {
        g_source_remove (m_repaint_timer);
        m_repaint_timer = 0;
    }
}

KDE_NO_EXPORT bool ViewArea::timeoutPaintEvent () {
    m_updaters = getFirstUpdater (m_updaters);
    if (m_updaters_enabled && m_updaters) {
            UpdateEvent event (m_updaters->node->document (), 0);
            propagateUpdatersEvent (m_updaters, &event);
    }
    if (!m_repaint_rect.isEmpty ())
        syncVisual (m_repaint_rect);
    if (m_updaters_enabled && m_updaters)
        return true;

    m_repaint_timer = 0;
    return false;
}

KDE_NO_EXPORT int ViewArea::width () const {
    int w, h;
    gdk_drawable_get_size (widget->window, &w, &h);
    return w;
}

KDE_NO_EXPORT int ViewArea::height () const {
    int w, h;
    gdk_drawable_get_size (widget->window, &w, &h);
    return h;
}

static void setXSelectInput (Display *dpy, Window wid, long mask) {
    XWindowAttributes attr;
    XGetWindowAttributes (dpy, wid, &attr);
    long m = mask | attr.your_event_mask;
    XSelectInput (dpy, wid, m);

    Window r, p, *c;
    unsigned int nr;
    if (XQueryTree (dpy, wid, &r, &p, &c, &nr)) {
        for (int i = 0; i < nr; ++i)
            setXSelectInput (dpy, c[i], mask);
        XFree (c);
    }
}

static bool isChildWindow (Display *dpy, Window p, Window c) {
    Window root = 0;
    Window parent = 0;
    Window *children;
    unsigned int nr;
    /*Gives 'BadWindow (invalid Window parameter)' when npp stops'
    while (p != c && XQueryTree (dpy, c, &root, &parent, &children, &nr)) {
        if (nr)
            XFree (children);
        if (parent == p || parent == root || c == root || !parent)
            break;
        c = parent;
    }
    return parent == p;*/
    bool found = false;
    if (XQueryTree (dpy, p, &root, &parent, &children, &nr)) {
        for (int i = 0; i < nr; ++i) {
            found = children[i] == c || isChildWindow (dpy, children[i], c);
            if (found)
                break;
        }
        if (nr)
            XFree (children);
    }
    return found;
}

GdkFilterReturn videoWidgetFilter (GdkXEvent *xe, GdkEvent *e, gpointer d) {
    XEvent *event = (XEvent *) xe;
    switch (event->xany.type) {
        case KeyPress:
        case KeyRelease: {
            Window v = GDK_DRAWABLE_XID (GDK_DRAWABLE (
                        ((ViewArea *)d)->videoWindow ()));
            if (event->xany.window == v ||
                    isChildWindow(event->xany.display, v, event->xkey.window)) {
                XKeyEvent keyevent = event->xkey;
                keyevent.window =
                    GDK_DRAWABLE_XID (GDK_DRAWABLE (((ViewArea *)d)->viewWidget ()->window));
                XSendEvent (event->xany.display, keyevent.window, true,
                        KeyPressMask, (XEvent *) &keyevent);
                return GDK_FILTER_REMOVE;
            }
            break;
        }
        case MapNotify: {
            Window v = GDK_DRAWABLE_XID (GDK_DRAWABLE (
                        ((ViewArea *)d)->videoWindow ()));
            if (!event->xmap.override_redirect &&
                 (event->xany.window == v ||
                  isChildWindow (event->xany.display, v, event->xmap.window))) {
                setXSelectInput (event->xany.display,
                        event->xmap.window,
                        KeyPressMask|ExposureMask|SubstructureNotifyMask);
            }
        }
    }
    return GDK_FILTER_CONTINUE;
}

KDE_NO_EXPORT void ViewArea::setEventFiltering (bool on) {
    if (on && !m_event_filter_on)
        gdk_window_add_filter (NULL, videoWidgetFilter, this);
    else if (!on && m_event_filter_on)
        gdk_window_remove_filter (NULL, videoWidgetFilter, this);
    m_event_filter_on = on;
}

KDE_NO_EXPORT void ViewArea::setOutputWidget (Application *app, GtkWidget * w) {
    widget = w;
    gdk_window_clear (w->window);
    if (surface->surface) {
        cairo_surface_destroy (surface->surface);
        surface->surface = 0L;
    }
    if (!video) {
        GdkWindowAttr attributes;
        attributes.window_type = GDK_WINDOW_CHILD;
        attributes.x = w->allocation.x;
        attributes.y = w->allocation.y;
        attributes.width = w->allocation.width;
        attributes.height = w->allocation.height;
        attributes.wclass = GDK_INPUT_OUTPUT;
        attributes.visual = gtk_widget_get_visual (w);
        attributes.colormap = gtk_widget_get_colormap (w);
        attributes.event_mask = gtk_widget_get_events (w) |
            GDK_SUBSTRUCTURE_MASK | GDK_EXPOSURE_MASK | GDK_KEY_PRESS_MASK;
        attributes.event_mask &= ~GDK_BUTTON_PRESS_MASK;
        //    ~(GDK_KEY_PRESS_MASK | GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
        gint mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
        video = gdk_window_new (w->window, &attributes, mask);
        gdk_window_show (video);
        setAudioVideoGeometry (IRect (-100, -100, 50, 50), 0L);

        debugLog() << "video window "  << (int)GDK_DRAWABLE_XID (GDK_DRAWABLE (video)) << " " << endl;

    } else {
        gdk_window_reparent (video, widget->window, m_av_geometry.x(), m_av_geometry.y());
    }
    gdk_window_focus (widget->window, CurrentTime);
    //resizeEvent ();
}

//------------------------------------------------------------------------------

void cb_view_area_button_press (GtkWidget *w, GdkEventButton *e, Application * app) {
        //debugLog () << "button press " << event->x << "," << event->y << endl;
    VIEWAREA(app)->mousePressEvent (int (e->x), int (e->y));
}

void cb_view_area_expose (GtkWidget *w, GdkEventExpose *e, Application *app) {
    debugLog () << "cb_view_area_expose " << e->area.x << "," << e->area.y << " " << e->area.width << "x" << e->area.height << endl;
    VIEWAREA(app)->paintEvent (
            IRect (e->area.x, e->area.y, e->area.width, e->area.height));
}

void cb_view_area_realized (GtkWidget *widget, Application * app) {
    gtk_widget_set_double_buffered (widget, false);
    if (widget == app->view_area_fullscreen) {
        Color c (0x101010);
        gdk_window_set_background (widget->window, (GdkColor *)(GColor)c);
    } else
        VIEWAREA(app)->setOutputWidget (app, widget);
}

gboolean cbViewAreaFullscreenMapped (GtkWidget *w, GdkEvent *, Application *a) {
    gtk_window_fullscreen (GTK_WINDOW (a->window_fullscreen));
    VIEWAREA(a)->setOutputWidget (a, a->view_area_fullscreen);
    cb_view_area_configure (w, 0L, a);
    return false;
}

void cbViewAreaFullscreenUnmapped (GtkWidget *w, Application *a) {
    VIEWAREA(a)->setOutputWidget (a, a->view_area);
    cb_view_area_configure (w, 0L, a);
}

gboolean cb_scheduledPaint (void * vp) {
    return static_cast <ViewArea *> (vp)->timeoutPaintEvent ();
}


//#ifdef gtk < 2.8

unsigned char * pixbufBits (GdkPixbuf *pixbuf, int & n_channels) {
    gint width = gdk_pixbuf_get_width (pixbuf);
    gint height = gdk_pixbuf_get_height (pixbuf);
    guchar *gdk_pixels = gdk_pixbuf_get_pixels (pixbuf);
    int gdk_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
    n_channels = gdk_pixbuf_get_n_channels (pixbuf);
    guchar *cairo_pixels, *head;
    static cairo_user_data_key_t key;
    int j;

    cairo_pixels = (guchar*)g_malloc (4 * width * height);
    head = cairo_pixels;

    for (j = height; j; j--) {
        guchar *p = gdk_pixels;
        guchar *q = cairo_pixels;

        if (n_channels == 3) {
            guchar *end = p + 3 * width;

            while (p < end) {
#if __BYTE_ORDER == __LITTLE_ENDIAN
                q[0] = p[2];
                q[1] = p[1];
                q[2] = p[0];
                q[3] = p[1];
#warning little ending
#else
                q[1] = p[0];
                q[2] = p[1];
                q[3] = p[2];
#warning big ending
#endif
                p += 3;
                q += 4;
            }
        } else {
            guchar *end = p + 4 * width;
            guint t1,t2,t3;

#define MULT(d,c,a,t) G_STMT_START { t = c * a + 0x7f; d = ((t >> 8) + t) >> 8; } G_STMT_END

            while (p < end) {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
                MULT(q[0], p[2], p[3], t1);
                MULT(q[1], p[1], p[3], t2);
                MULT(q[2], p[0], p[3], t3);
                q[3] = p[3];
#else
                q[0] = p[3];
                MULT(q[1], p[0], p[3], t1);
                MULT(q[2], p[1], p[3], t2);
                MULT(q[3], p[2], p[3], t3);
#endif

              p += 4;
              q += 4;
            }

#undef MULT
        }

        gdk_pixels += gdk_rowstride;
        cairo_pixels += 4 * width;
    }
    return head;
}

/*
void
gdk_cairo_set_source_pixbuf (cairo_t   *cr,
                             GdkPixbuf *pixbuf,
                             double     pixbuf_x,
                             double     pixbuf_y)
{
  gint width = gdk_pixbuf_get_width (pixbuf);
  gint height = gdk_pixbuf_get_height (pixbuf);
  guchar *gdk_pixels = gdk_pixbuf_get_pixels (pixbuf);
  int gdk_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
  int n_channels = gdk_pixbuf_get_n_channels (pixbuf);
  debugLog() << "gdk_cairo_set_source_pixbuf pixbuf:" << width << "x" << height << " rowstride:" << gdk_rowstride << " channels:" << n_channels << endl;
  guchar *cairo_pixels;
  cairo_format_t format;
  cairo_surface_t *surface;
  static cairo_user_data_key_t key;
  int j;

  if (n_channels == 3)
    format = CAIRO_FORMAT_RGB24;
  else
    format = CAIRO_FORMAT_ARGB32;

  cairo_pixels = (guchar*)g_malloc (4 * width * height);
  surface = cairo_image_surface_create_for_data ((unsigned char *)cairo_pixels,
                                                 format,
                                                 width, height, 4 * width);
  cairo_surface_set_user_data (surface, &key,
                               cairo_pixels, (cairo_destroy_func_t)g_free);

  for (j = height; j; j--)
    {
      guchar *p = gdk_pixels;
      guchar *q = cairo_pixels;

      if (n_channels == 3)
        {
          guchar *end = p + 3 * width;
          
          while (p < end)
            {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
              q[0] = p[2];
              q[1] = p[1];
              q[2] = p[0];
#else     
              q[1] = p[0];
              q[2] = p[1];
              q[3] = p[2];
#endif
              p += 3;
              q += 4;
            }
        }
      else
        {
          guchar *end = p + 4 * width;
          guint t1,t2,t3;
            
#define MULT(d,c,a,t) G_STMT_START { t = c * a + 0x7f; d = ((t >> 8) + t) >> 8; } G_STMT_END

          while (p < end)
            {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
              MULT(q[0], p[2], p[3], t1);
              MULT(q[1], p[1], p[3], t2);
              MULT(q[2], p[0], p[3], t3);
              q[3] = p[3];
#else     
              q[0] = p[3];
              MULT(q[1], p[0], p[3], t1);
              MULT(q[2], p[1], p[3], t2);
              MULT(q[3], p[2], p[3], t3);
#endif
              
              p += 4;
              q += 4;
            }
          
#undef MULT
        }

      gdk_pixels += gdk_rowstride;
      cairo_pixels += 4 * width;
    }

  cairo_set_source_surface (cr, surface, pixbuf_x, pixbuf_y);
  cairo_surface_destroy (surface);
}
*/
